import logging
import dataclasses
from typing import Type
from .models import User, App
from pydantic import BaseModel, Field, ConfigDict, field_validator
import datetime


logger = logging.getLogger(__name__)


class JWTToken(BaseModel):
    """A JWT token

    This is a pydantic model that represents a JWT token.
    It is used to validate the token and to extract information from it.
    The token is decoded using the `decode_token` function.

    """
    model_config = ConfigDict(extra="forbid")

    sub: str
    """A unique identifier for the user (is unique for the issuer)"""
    iss: str
    """The issuer of the token"""
    
    
    exp: datetime.datetime
    """The expiration time of the token"""
    
    
    client_id: str
    """The client_id of the app that requested the token"""
    preferred_username: str
    """The username of the user"""
    roles: list[str]
    """The roles of the user"""
    scope: str
    """The scope of the token"""
    
    iat: datetime.datetime
    """The issued at time of the token"""

    aud: str | None = None
    """The audience of the token"""
    
    jti: str | None = None
    """The unique identifier for the token"""

    raw: str
    """ The raw original token string """

    @field_validator("sub", mode="before")
    def sub_to_username(cls: Type["JWTToken"], v: str) -> str:
        """Convert the sub to a username compatible string"""
        if isinstance(v, int):
            return str(v)
        return v
    
    @field_validator("iat", mode="before")
    def iat_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the iat to a datetime object"""
        if v is None:
            return None
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v
    
    
    @field_validator("exp", mode="before")
    def exp_to_datetime(cls: Type["JWTToken"], v: int) -> datetime.datetime:
        """Convert the exp to a datetime object"""
        if isinstance(v, int):
            return datetime.datetime.fromtimestamp(v)
        return v

    @property
    def changed_hash(self) -> str:
        """A hash that changes when the user changes"""
        return str(hash(self.sub + self.preferred_username + " ".join(self.roles)))

    @property
    def scopes(self) -> list[str]:
        """The scopes of the token. Each scope is a string separated by a space"""
        return self.scope.split(" ")



class StaticToken(JWTToken):
    """A static JWT token"""

    sub: str
    iss: str
    iat: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now()
    )
    exp: datetime.datetime = Field(
        default_factory=lambda: datetime.datetime.now() + datetime.timedelta(days=1)
    )
    client_id: str = "static"
    preferred_username: str = "static_user"
    scope: str = "openid profile email"
    roles: list[str] = Field(default_factory=lambda: ["static"])
    raw: str = Field(default_factory=lambda: "static_token")
    
    
    


class AuthentikateSettings(BaseModel):
    """The settings for authentikate

    This is a pydantic model that represents the settings for authentikate.
    It is used to configure the library.
    """
    allowed_audiences: list[str] = Field(default_factory=lambda: ["rekuest"])
    algorithms: list[str]
    public_key: str
    force_client: bool
    allow_imitate: bool
    imitate_headers: list[str] = Field(default_factory=lambda: ["X-Imitate-User"])
    authorization_headers: list[str] = Field(
        default_factory=lambda: [
            "Authorization",
            "X-Authorization",
            "AUTHORIZATION",
            "authorization",
        ]
    )
    imitate_permission: str = "authentikate.imitate"
    static_tokens: dict[str, StaticToken] = Field(default_factory=dict)
    """A map of static tokens to their decoded values. Should only be used in tests."""


@dataclasses.dataclass
class Auth:
    """
    Mimics the structure of `AbstractAccessToken` so you can use standard
    Django Oauth Toolkit permissions like `TokenHasScope`.
    """

    token: JWTToken
    user: User
    app: App

    def is_valid(self, scopes: list[str] | None = None) -> bool:
        """
        Check if the token is valid


        Parameters
        ----------
        scopes : list[str], optional
            The scopes to check, by default None

        Returns
        -------
        bool
            Whether the token is valid

        """
        return not self.is_expired() and self.has_scopes(scopes or [])

    def is_expired(self) -> bool:
        """
        Check if the token is expired

        As we are using JWT tokens, we do not need to check the expiration time
        as the token is already checked for expiration when it is decoded.

        Returns
        -------
        bool
            Whether the token is expired or not (will always return False)
        """
        # Token expiration is already checked
        return False

    def has_scopes(self, scopes: list[str]) -> bool:
        """Does the token have the required scopes?

        Check if the token has the required scopes, if no scopes are provided
        it will return True.

        Parameters
        ----------
        scopes : list[str]
            The scopes to check

        Returns
        -------
        bool
            Does the token have the required scopes?
        """

        provided_scopes = set(self.token.scopes)
        resource_scopes = set(scopes)

        return resource_scopes.issubset(provided_scopes)



from django.db import models  # Create your models here.
from django.contrib.auth.models import AbstractUser


class User(AbstractUser):
    """A reflection on the real User"""

    sub = models.CharField(max_length=1000, null=True, blank=True)
    iss = models.CharField(max_length=1000, null=True, blank=True)
    changed_hash = models.CharField(max_length=1000, null=True, blank=True)

    class Meta:
        """Meta class for User"""

        constraints = [
            models.UniqueConstraint(
                fields=["sub", "iss"],
                condition=models.Q(sub__isnull=False, iss__isnull=False),
                name="unique_sub_iss_if_both_not_null",
            )
        ]
        permissions = [("imitate", "Can imitate me")]


class App(models.Model):
    """An Oauth2 App

    An Oauth2 App is a model to represent an Oauth2 app that is
    registered when a JWT token is authenticated. It retrieves
    the client_id from the token and uses it to create a new
    app or retrieve an existing app. This allows for the grouping
    of users by app.

    """
    iss = models.CharField(max_length=2000, null=True, blank=True)
    client_id = models.CharField(unique=True, max_length=2000)
    name = models.CharField(max_length=2000, null=True, blank=True)

    class Meta:
        """Meta class for App"""

        unique_together = ("iss", "client_id")

    def __str__(self) -> str:
        """String representation of App"""
        return f"{self.name}"



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
import os
from authentikate.base_models import AuthentikateSettings
from typing import Optional

cached_settings: Optional[AuthentikateSettings] = None


def prepare_settings() -> AuthentikateSettings:
    """Prepare the settings

    Prepare the settings for authentikate from django_settings.
    This function will raise a ImproperlyConfigured exception if the settings are
    not correct.

    Returns
    -------
    AuthentikateSettings
        The settings

    Raises
    ------
    ImproperlyConfigured
        When the settings are not correct
    """

    try:
        user = settings.AUTH_USER_MODEL
        if user != "authentikate.User":
            raise ImproperlyConfigured(
                "AUTH_USER_MODEL must be authentikate.User in order to use authentikate"
            )
    except AttributeError:
        raise ImproperlyConfigured(
            "AUTH_USER_MODEL must be authentikate.User in order to use authentikate"
        )

    try:
        group = settings.AUTHENTIKATE
    except AttributeError:
        raise ImproperlyConfigured("Missing setting AUTHENTIKATE")

    try:
        algorithms = [group["KEY_TYPE"]]

        public_key = group.get("PUBLIC_KEY", None)
        allow_imitate = group.get("ALLOW_IMITATE", True)
        imitation_headers = group.get("IMITATION_HEADERS", None)
        imitate_permission = group.get("IMITATE_PERMISSION", None)
        authorization_headers = group.get("AUTHORIZATION_HEADERS", None)
        static_tokens = group.get("STATIC_TOKENS", {})

        if not public_key:
            pem_file: str = group.get("PUBLIC_KEY_PEM_FILE", None)  # type: ignore
            if not pem_file:
                raise ImproperlyConfigured(
                    "Missing setting in AUTHENTIKAE: PUBLIC_KEY_PEM_FILE (path to public_key.pem) or PUBLIC_KEY (string of public key)"
                )

            try:
                base_dir = settings.BASE_DIR
            except AttributeError:
                raise ImproperlyConfigured("Missing setting AUTHENTIKATE")

            try:
                path = os.path.join(base_dir, pem_file)

                with open(path, "rb") as f:
                    public_key = f.read()

            except FileNotFoundError:
                raise ImproperlyConfigured(f"Pem File not found: {path}")

        force_client = group.get("FORCE_CLIENT", False)

        return AuthentikateSettings(
            algorithms=algorithms,
            public_key=public_key,
            force_client=force_client,
            imitation_headers=imitation_headers,
            authorization_headers=authorization_headers,
            allow_imitate=allow_imitate,
            imitate_permission=imitate_permission,
            static_tokens=static_tokens,
        )

    except KeyError:
        raise ImproperlyConfigured(
            "Missing setting AUTHENTIKATE KEY_TYPE or AUTHENTIKATE PUBLIC_KEY"
        )


def get_settings() -> AuthentikateSettings:
    """Get the settings

    Returns
    -------

    AuthentikateSettings
        The settings
    """
    global cached_settings
    if not cached_settings:
        cached_settings = prepare_settings()
    return cached_settings



from authentikate.base_models import Auth, AuthentikateSettings
from authentikate.models import User


def imitate_user(auth: Auth, imitate_id: str, settings: AuthentikateSettings) -> Auth:
    """Imitate a user

    Parameters
    ----------
    auth : Auth
        The auth context to imitate

    imitate_id : str
        The user to imitate needs to folow the format sub@iss

    settings : AuthentikateSettings
        The settings to use

    Returns
    -------
    Auth
        The new auth context
    """
    if "@" not in imitate_id:
        raise ValueError("Imitate ID must be in the format sub@iss")
    
    
    sub, iss = imitate_id.split("@")
    user = User.objects.get(sub=sub, iss=iss)

    if auth.user.has_perm(settings.imitate_permission, user):
        auth.user = user
        return auth

    else:
        raise PermissionError("User does not have permission to imitate this user")



from authentikate.errors import AuthentikatePermissionDenied
import time
from django.contrib.auth.models import Group
from authentikate import base_models, models
import logging


logger = logging.getLogger(__name__)


def token_to_username(token: base_models.JWTToken) -> str:
    """Convert a JWT token to a username

    Parameters
    ----------
    token : structs.JWTToken
        The token to convert

    Returns
    -------

    str
        The username



    """
    return f"{token.iss}_{token.sub}"


def set_user_groups(user: models.User, roles: list[str]) -> None:
    """Add a list of roles to a user

    Roles are added as groups

    Parameters
    ----------
    user : models.User
        The user to add the roles to
    roles : list[str]
        The roles to add
    """
    for role in roles:
        g, _ = Group.objects.get_or_create(name=role)
        user.groups.add(g)


def expand_token(token: base_models.JWTToken, force_client: bool = True) -> base_models.Auth:
    """Expand a JWT token into an Auth context

    Parameters
    ----------
    token : structs.JWTToken
        The token to expand
    force_client : bool, optional
        Whether to force the client to be present, by default True

    Returns
    -------
    structs.Auth
        The expanded token
    """

    if token.sub is None:
        raise AuthentikatePermissionDenied("Missing sub parameter in JWT token")

    if token.iss is None:
        raise AuthentikatePermissionDenied("Missing iss parameter in JWT token")

    if token.exp is None:
        raise AuthentikatePermissionDenied("Missing exp parameter in JWT token")

    # Check if token is expired
    if token.exp.timestamp() < time.time():
        raise AuthentikatePermissionDenied("Token has expired")

    if token.client_id is None:
        if force_client:
            raise AuthentikatePermissionDenied(
                "Missing client_id parameter in JWT token"
            )

    try:
        if token.client_id is None:
            app = None
        else:
            app, _ = models.App.objects.get_or_create(
                client_id=token.client_id, iss=token.iss
            )

        user = models.User.objects.get(sub=token.sub, iss=token.iss)
        if user.changed_hash != token.changed_hash:
            # User has changed, update the user object
            user.first_name = token.preferred_username
            user.changed_hash = token.changed_hash
            set_user_groups(user, token.roles)
            user.save()

    except models.User.DoesNotExist:
        preexisting_user = models.User.objects.filter(
            username=token.preferred_username
        ).first()

        user = models.User(
            sub=token.sub,
            username=token_to_username(token)
            if preexisting_user
            else token.preferred_username,
            iss=token.iss,
            first_name=token.preferred_username,
        )
        user.set_unusable_password()
        user.save()
        user.first_name = token.preferred_username
        user.changed_hash = token.changed_hash
        set_user_groups(user, token.roles)
        user.save()
    except Exception as e:
        logger.error(f"Error while authenticating: {e}", exc_info=True)
        raise AuthentikatePermissionDenied(f"Error while authenticating: {e}")

    return base_models.Auth(
        token=token,
        user=user,
        app=app,
    )
